#!/usr/bin/python3
from datetime import datetime, timedelta
from traceback import print_exc
import traceback
from ltud_pb2 import Empty
from ltud_pb2_grpc import LTUdStub
import grpc
from pcp.pmda import PMDA

class LTUDObject:
    inactivity_limit = 10
    inactivity_counter = 0
    counters = []
    last_counters = []
    rates = []
    names = []
    timestamp = 0.0
    last_timestamp = 0.0
    updated = datetime.now() - timedelta(seconds=3600)
    host = ""
    port = ""
    timeout = -1
    heartbeat = 0
    disabled = False
    name = ""
    description = ""
    channel = None
    stub = None
    options = None

    def __init__(self, host, port, name, description, options, timeout = 2):
        if options.debug:
            PMDA.dbg("[%s] LTUdObject Initialisation" % name)
        self.options = options
        self.host = host
        self.port = port
        self.timeout = timeout
        self.updated = datetime.now() - timedelta(seconds=3600)
        self.name = name
        self.description = description
        PMDA.log("[%s] LTUdObject __init__ - connecting to gRPC server %s:%s" % (self.name, self.host, self.port))
        try:
            self.channel = grpc.insecure_channel(host + ":" + port)
            self.stub = LTUdStub(self.channel)
        except Exception as e:
            PMDA.err(str(e))
            traceback.print_exc()
            return
        if options.names:
            self.update_grpc()
            for n in self.names:
                if len(n) > 0:
                    print("cts." + name + "." + n)
            PMDA.dbg("Updated: " + str(self.updated))
            return
    
    def update(self):
        if self.options.debug:
            PMDA.dbg("[%s] LTUdObject update" % self.name)
        # PMDA.dbg("NOW: " + str(datetime.now()))
        # PMDA.dbg("UPDATED: " + str(self.updated))
        # PMDA.dbg("Calculated: " + str((datetime.now() - self.updated).total_seconds()))
        # PMDA.dbg("TIMEOUT: " + str(self.timeout))
        if (datetime.now() - self.updated).total_seconds() > self.timeout:
            if self.options.debug:
                PMDA.dbg("[%s] LTUdObject update - Cache invalid, need to contact gRPC server" % self.name)
            if self.disabled:
                self.inactivity_counter += 1
            self.update_grpc()
            # PMDA.dbg("Last timestamp: " + str(int(self.last_timestamp)) + str(int(self.last_timestamp) == 0))
            if int(self.last_timestamp) == 0:
                # PMDA.dbg("Will update once more")
                # PMDA.dbg("Self updated before: " + str(self.updated))
                self.updated = datetime.now() - timedelta(seconds=2 * self.timeout)
                # PMDA.dbg("Self updated: " + str(self.updated))
                # PMDA.dbg("Diff: " + str((datetime.now() - self.updated).total_seconds()))
            return [self.rates, 1]
        return [self.rates, 1]

    def update_grpc(self):
        if self.disabled:
            if self.options.debug:
                PMDA.dbg("[%s] LTUdObject update_grpc - this object is disabled [%d / %d]" % (self.name, self.inactivity_counter, self.inactivity_limit))
            if self.inactivity_counter >= self.inactivity_limit:
                PMDA.dbg("[%s] LTUdObject update_grpc - inactivity reached the limit" % self.name)
                self.inactivity_counter = 0
                self.disabled = False
                try:
                    PMDA.dbg("[%s] LTUdObject update_grpc - reconnecting to the gRPC server %s:%s" % (self.name, self.host, self.port))
                    self.channel = grpc.insecure_channel(self.host + ":" + self.port)
                    self.stub = LTUdStub(self.channel)
                except Exception as e:
                    PMDA.err(str(e))
                    self.disabled = True
            return
        try:
            if len(self.names) == 0:
                if self.options.debug:
                    PMDA.dbg("[%s] LTUdObject update_grpc - this object has no names yet" % self.name)
                self.names = self.stub.OutCntsGetNames(Empty()).names
                if self.options.debug:
                    for idx, n in enumerate(self.names):
                        PMDA.dbg("[%d] cts.%s.%s" % (idx, self.name, n))
                return 0
            raw_counters = self.stub.OutCntsGet(Empty())
            self.last_timestamp = self.timestamp
            self.timestamp = raw_counters.timestamp
            # PMDA.dbg("Last Timestamp: " + str(self.last_timestamp))
            # PMDA.dbg("Timestamp: " + str(self.timestamp))
            self.last_counters = self.counters
            self.counters = raw_counters.outcnts
            self.calculate_rate()
            self.updated = datetime.now()
            return 0
        # except grpc._channel._InactiveRpcError as e:
        #     PMDA.err(str(e))
        #     PMDA.err("Ahoj svet")
        #     return 0
        except Exception as e:
            PMDA.err(str(e))
            # print_exc()
            self.updated = datetime.now()
            self.disabled = True
            return 0
    
    def calculate_rate(self):
        if len(self.rates) == 0:
            self.rates = self.counters
        diff_timestamp = self.timestamp - self.last_timestamp
        if self.last_timestamp == 0 or len(self.last_counters) == 0 or diff_timestamp == 0:
            # TODO if timestamp is 0, we need to send to PCP that value is N/A
            for index, counter in enumerate(self.counters):
                # self.counters[index] = 0
                self.rates[index] = 0
            return
        for index, counter in enumerate(self.counters):
            # self.counters[index] = int((counter - self.old_counters[index]) / self.timestamp)
            try:
                self.rates[index] = int((counter - self.last_counters[index]) / diff_timestamp)
                PMDA.dbg(str(self.rates[index]))
            except ValueError as e:
                # When last counter value is 0, return rate 0
                self.rates[index] = 0